unit FormMain;
{ TODO : Implement Damsels }
{ TODO : Implement WanderingMine }
{ TODO : Implement Sound }

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  Grids, StdCtrls, Buttons, Arrow, fgl, contnrs, Types;

type

  TContent = (Empty, Mine, Damsel, Target, AlreadyEntered);
  TPos = class
    Col : Integer;
    Row : Integer;
  end;

  TMovesList = specialize TFPGObjectList<TPos>;

  TLevel = record
    Points : Integer;
    Mines : Integer;
    HaveDamsels : Boolean;
    WanderingMine : Boolean;
  end;

  { TMainForm }

  TMainForm = class(TForm)
    ArrowGoLeft: TArrow;
    ArrowGoUp: TArrow;
    ArrowGoRight: TArrow;
    ArrowGoDown: TArrow;
    ReplayButton: TButton;
    NextLevelButton: TButton;
    OKButton: TButton;
    ButtonNewGame: TButton;
    ButtonExit: TButton;
    ControlPanel: TPanel;
    MessageText: TMemo;
    MessagePanel: TPanel;
    Pictures: TImageList;
    Pitch: TDrawGrid;
    MinesLabel: TLabel;
    PointsLabel: TLabel;
    PointsTitleLabel: TLabel;
    MinesTitleLabel: TLabel;
    LevelLabel: TLabel;
    LevelTitleLable: TLabel;
    HeadingStaticText: TStaticText;
    ReplayTimer: TTimer;
    procedure ArrowGoDownClick(Sender: TObject);
    procedure ArrowGoLeftClick(Sender: TObject);
    procedure ArrowGoRightClick(Sender: TObject);
    procedure ArrowGoUpClick(Sender: TObject);
    procedure ButtonExitClick(Sender: TObject);
    procedure ButtonNewGameClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure InitializePitch;
    procedure NextLevelButtonClick(Sender: TObject);
    procedure OKButtonClick(Sender: TObject);
    procedure PitchDrawCell(Sender: TObject; aCol, aRow: integer;
      aRect: TRect; aState: TGridDrawState);
    procedure PitchMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: integer);
    procedure PitchMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure ReplayButtonClick(Sender: TObject);
    procedure ReplayTimerTimer(Sender: TObject);
  private
    { private declarations }
    FPlayer: TBitmap;
    FMine: TBitmap;
    FTarget: TBitmap;
    FExplosion : TBitmap;
    FAlreadyEntered : TBitmap;
    FDamsel : TBitmap;
    FNumberOfCarriedDamsels : Integer;
    FLevelIndex: Integer;
    FLevels : array[1..10] of TLevel;
    FMoves : TMovesList;
    FMovesIndex : Integer;
    FPitchContent: array of array of TContent;
    FGameLost: boolean;
    FReplayMode : boolean;
    FPoints : Integer;
    procedure CheckMines;
    procedure InitializeLevels;
    procedure LevelFinished;
    procedure ShowAlreadyEnteredFields(const Pos: TPos; const Idx : Integer);
    procedure ShowPoints;
    procedure WinGame;
    procedure GotDamsel;
  public
    { public declarations }
  end;


var
  MainForm: TMainForm;

  FLevel: integer;

implementation

{$R *.lfm}

{ TMainForm }

procedure TMainForm.FormCreate(Sender: TObject);
begin
  SetLength(FPitchContent, Pitch.ColCount, Pitch.RowCount);
  FMoves := TMovesList.Create;
  FPlayer := TBitmap.Create;
  FTarget := TBitmap.Create;
  FMine := TBitmap.Create;
  FAlreadyEntered := TBitmap.Create;
  FDamsel := TBitmap.Create;
  FExplosion := TBitmap.Create;
  Pictures.GetBitmap(0, FPlayer);
  Pictures.GetBitmap(1, FTarget);
  Pictures.GetBitmap(2, FMine);
  Pictures.GetBitmap(3, FExplosion);
  Pictures.GetBitmap(4, FAlreadyEntered);
  Pictures.GetBitmap(5, FDamsel);
  FLevelIndex := 0; // Index will be increased during initialising of pitch
  InitializeLevels;
  FReplayMode:= False;

end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FPlayer.Free;
  FTarget.Free;
  FMine.Free;
  FExplosion.Free;
  FAlreadyEntered.Free;
  FMoves.Free;
  FDamsel.Free;
  Pictures.Free;
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  Pitch.SetFocus;
end;

procedure TMainForm.InitializePitch;
var
  i, j: integer;
  X, Y: integer;
  NumberOfMines : Integer;
begin
  { Define Playerposition}
  Pitch.Col := Pitch.ColCount div 2;
  Pitch.Row := Pitch.RowCount - 1;
  { Reset Settings}
  FGameLost := False;
  Pitch.Enabled:= True;
  MessagePanel.Visible:= False;
  OKButton.Visible:= False;
  NextLevelButton.Visible:= True;
  FMoves.Clear;
  FNumberOfCarriedDamsels:= 0;
  { Increase Gamelevel}
  Inc(FLevelIndex);
  if FLevelIndex > 10 then
    WinGame;
  LevelLabel.Caption:= IntToStr(FLevelIndex);
  {Clear FPitchContent}
  for i := 0 to Pitch.ColCount - 1 do
    for j := 0 to Pitch.RowCount - 1 do
      FPitchContent[i, j] := Empty;
  Pitch.Invalidate;


  { Define Targetposition}
  for i := ((Pitch.ColCount div 2) - 2) to ((Pitch.ColCount div 2) + 2) do
  begin
    FPitchContent[i, 0] := Target;
  end;
  { Define Damselpositions }
  if FLevels[FLevelIndex].HaveDamsels then
  begin
    Y := Pitch.RowCount div 3;
    // Position of first Damsel
    X := Pitch.ColCount div 4;
    FPitchContent[X, Y] := Damsel;
    // Position of second Damsel
    X := Pitch.ColCount - (Pitch.ColCount div 4);
    FPitchContent[X, Y] := Damsel;
  end;
  { Mining}
  NumberOfMines := FLevels[FLevelIndex].Mines;
  Randomize;
  while NumberOfMines > 0 do
  begin
    X := Random(Pitch.ColCount);
    Y := Random(Pitch.RowCount);
    { No Mines in the first and last row!}
    if (Y > 0) and (Y < Pitch.RowCount - 1) then
    begin
      {put down mine, if it is an empty place}
      if FPitchContent[X, Y] = Empty then
      begin
        FPitchContent[X, Y] := Mine;
        Dec(NumberOfMines);
      end;
    end;
  end;
  Pitch.SetFocus;
end;

procedure TMainForm.NextLevelButtonClick(Sender: TObject);
begin
  InitializePitch;
end;

procedure TMainForm.OKButtonClick(Sender: TObject);
begin
  MessagePanel.Visible:= False;
end;

procedure TMainForm.PitchDrawCell(Sender: TObject; aCol, aRow: integer;
  aRect: TRect; aState: TGridDrawState);
var
  Pos : TPos;
begin

    if (aCol = Pitch.Col) and (aRow = Pitch.Row) then
    Case FPitchContent[aCol, aRow] of
      Mine:
        begin
          FGameLost:= True;
          Pitch.Canvas.Draw(aRect.Left, aRect.Top, FExplosion);
          Pitch.Enabled:= False;
        end;
      Target:
        if (MessagePanel.Visible = False) and (FReplayMode = False) then
          LevelFinished;
      Damsel:
        begin
          GotDamsel;
          FPitchContent[aCol, aRow] := AlreadyEntered;
        end;
      else
        begin
          Pitch.Canvas.Draw(aRect.Left, aRect.Top, FPlayer);
          Pos := TPos.Create;
          if not FReplayMode then
            begin
              Pos.Col:= aCol;
              Pos.Row:= aRow;
              FMoves.Add(Pos);
            end;
          if FMoves.Count > 0 then
            ShowAlreadyEnteredFields(Pos, FMoves.Count-1);
          CheckMines;
        end;
      end

    else if FPitchContent[aCol, aRow] = Mine then
      begin
      if FReplayMode then
        Pitch.Canvas.Draw(aRect.Left, aRect.Top, FMine)
      end
    else if FPitchContent[aCol, aRow] = Target then
      Pitch.Canvas.Draw(aRect.Left, aRect.Top, FTarget)
    else if FPitchContent[aCol, aRow] = AlreadyEntered then
      Pitch.Canvas.Draw(aRect.Left, aRect.Top, FAlreadyEntered)
    else if FPitchContent[aCol, aRow] = Damsel then
      Pitch.Canvas.Draw(aRect.Left, aRect.Top, FDamsel);

end;

procedure TMainForm.PitchMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: integer);
begin
  Abort;
end;

procedure TMainForm.PitchMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
begin
  Abort;
end;

procedure TMainForm.ReplayButtonClick(Sender: TObject);
var
  i, j : Integer;
begin
  FReplayMode:= True;
  FMovesIndex:= 0;
  {Set Player to startposition}
  Pitch.Col := Pitch.ColCount div 2;
  Pitch.Row := Pitch.RowCount - 1;
  {delete AlreadyEntered-steps}
  for i := 0 to Pitch.ColCount - 1 do
    for j := 0 to Pitch.RowCount - 1 do
    if FPitchContent[i, j] = AlreadyEntered then
      FPitchContent[i, j] := Empty;
  Pitch.Invalidate;
  MessagePanel.Visible:= False;
  ReplayTimer.Enabled:= True;
end;

procedure TMainForm.ReplayTimerTimer(Sender: TObject);
var
  Pos : TPos;
begin
  if FMoves.Count > FMovesIndex then
    begin
      Pos := TPos.Create;
      Pitch.Col:= FMoves[FMovesIndex].Col;
      Pitch.Row:= FMoves[FMovesIndex].Row;
      Pos.Col:= Pitch.Col;
      Pos.Row:= Pitch.Row;
      ShowAlreadyEnteredFields(Pos, FMovesIndex);
      Inc(FMovesIndex);
      Pos.Free;
    end
  else
    begin
      ReplayTimer.Enabled:= False;
      FReplayMode:= False;
      MessagePanel.Visible:= True;
    end;
end;

procedure TMainForm.CheckMines;
var
  NumberOfMines: integer;
begin
  NumberOfMines := 0;
  if Pitch.Col > 0 then
    if FPitchContent[Pitch.Col - 1, Pitch.Row] = Mine then
      Inc(NumberOfMines);
  if Pitch.Col < Pitch.ColCount - 1 then
    if FPitchContent[Pitch.Col + 1, Pitch.Row] = Mine then
      Inc(NumberOfMines);
  if Pitch.Row > 0 then
    if FPitchContent[Pitch.Col, Pitch.Row - 1] = Mine then
      Inc(NumberOfMines);
  if Pitch.Row < Pitch.RowCount - 1 then
    if FPitchContent[Pitch.Col, Pitch.Row + 1] = Mine then
      Inc(NumberOfMines);
  MinesLabel.Caption := IntToStr(NumberOfMines);
  case NumberOfMines of
    0: MinesLabel.Font.Color := clBlack;
    1: MinesLabel.Font.Color := clOlive;
    2: MinesLabel.Font.Color := clMaroon;
    3: MinesLabel.Font.Color := clRed;
  end;
end;

procedure TMainForm.InitializeLevels;
begin
  With FLevels[1] do
  begin
    Points:= 150;
    Mines := 80;
    HaveDamsels := True;
    WanderingMine := False;
  end;
  With FLevels[2] do
  begin
    Points:= 200;
    Mines := 90;
    HaveDamsels := False;
    WanderingMine := False;
  end;
  With FLevels[3] do
  begin
    Points:= 250;
    Mines := 100;
    HaveDamsels := False;
    WanderingMine := False;
  end;
  With FLevels[4] do
  begin
    Points:= 300;
    Mines := 110;
    HaveDamsels := True;
    WanderingMine := False;
  end;
  With FLevels[5] do
  begin
    Points:= 350;
    Mines := 120;
    HaveDamsels := True;
    WanderingMine := False;
  end;
  With FLevels[6] do
  begin
    Points:= 400;
    Mines := 130;
    HaveDamsels := True;
    WanderingMine := True;
  end;
  With FLevels[7] do
  begin
    Points:= 450;
    Mines := 140;
    HaveDamsels := True;
    WanderingMine := True;
  end;
  With FLevels[8] do
  begin
    Points:= 500;
    Mines := 150;
    HaveDamsels := True;
    WanderingMine := True;
  end;
  With FLevels[9] do
  begin
    Points:= 750;
    Mines := 170;
    HaveDamsels := True;
    WanderingMine := True;
  end;
  With FLevels[10] do
  begin
    Points:= 1000;
    Mines := 200;
    HaveDamsels := True;
    WanderingMine := True;
  end;
end;

procedure TMainForm.LevelFinished;
var
  PointsInThisLevel : Integer;
begin
  Pitch.Enabled:= False;
  PointsInThisLevel := FLevels[FLevelIndex].Points - FMoves.Count;
  if PointsInThisLevel < 25 then
    PointsInThisLevel:= 25;
  FPoints:= FPoints + PointsInThisLevel;
  if FNumberOfCarriedDamsels > 0 then
     PointsInThisLevel:= PointsInThisLevel + (FNumberOfCarriedDamsels * 50);
  ShowPoints;
  MessageText.Clear;
  MessageText.Lines.Add('Congratulations!');
  MessageText.Lines.Add('You finished Level ' + IntToStr(FLevelIndex));
  if FNumberOfCarriedDamsels > 0 then
  begin
    MessageText.Lines.Add('You rescued '+IntToStr(FNumberOfCarriedDamsels)+
    ' Damsels');
    MessageText.Lines.Add('(50 extra points for each damsel)');
  end;

  MessageText.Lines.Add('You receive ' + IntToStr(PointsInThisLevel)
    + ' Points');
  MessagePanel.Visible:= True;

end;

procedure TMainForm.ShowAlreadyEnteredFields(const Pos: TPos; const Idx : Integer);
begin
  Pos.Col:= FMoves[Idx].Col;
  Pos.Row:= FMoves[Idx].Row;
  FPitchContent[Pos.Col, Pos.Row] := AlreadyEntered;
end;

procedure TMainForm.ShowPoints;
begin
  PointsLabel.Caption:= IntToStr(FPoints);
end;

procedure TMainForm.WinGame;
begin
  MessageText.Clear;
  MessageText.Lines.Add('You got it!!!');
  MessageText.Lines.Add('');
  MessageText.Lines.Add('You win this game - congratulations!');
  MessageText.Lines.Add('');
  MessageText.Lines.Add('You have got ' + IntToStr(FPoints) + ' Points');
  OKButton.Visible:= True;
  NextLevelButton.Visible:= False;
  Pitch.Enabled:= False;
end;

procedure TMainForm.GotDamsel;
begin
  FPoints := FPoints + 50;
  Inc(FNumberOfCarriedDamsels);
  ShowPoints;
end;



procedure TMainForm.ArrowGoUpClick(Sender: TObject);
begin
   if Pitch.Row > 0 then
     Pitch.Row:= Pitch.Row-1;
end;

procedure TMainForm.ArrowGoDownClick(Sender: TObject);
begin
  if Pitch.Row < Pitch.RowCount-1 then
    Pitch.Row:= Pitch.Row+1;
end;

procedure TMainForm.ArrowGoLeftClick(Sender: TObject);
begin
  if Pitch.Col > 0 then
    Pitch.Col:= Pitch.Col-1;
end;

procedure TMainForm.ArrowGoRightClick(Sender: TObject);
begin
  if Pitch.Col < Pitch.ColCount-1 then
    Pitch.Col:= Pitch.Col+1;
end;

procedure TMainForm.ButtonExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.ButtonNewGameClick(Sender: TObject);
begin
  FLevelIndex:= 0;
  FPoints:= 0;
  LevelLabel.Caption:=IntToStr(FLevelIndex);
  PointsLabel.Caption:=IntToStr(FPoints);
  InitializePitch;

end;

procedure TMainForm.FormActivate(Sender: TObject);
begin
  ButtonNewGameClick(self);
end;

end.

